#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###################################################################################
#                                                                                 #
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/               #
# Contact: beat.support@idiap.ch                                                  #
#                                                                                 #
# Redistribution and use in source and binary forms, with or without              #
# modification, are permitted provided that the following conditions are met:     #
#                                                                                 #
# 1. Redistributions of source code must retain the above copyright notice, this  #
# list of conditions and the following disclaimer.                                #
#                                                                                 #
# 2. Redistributions in binary form must reproduce the above copyright notice,    #
# this list of conditions and the following disclaimer in the documentation       #
# and/or other materials provided with the distribution.                          #
#                                                                                 #
# 3. Neither the name of the copyright holder nor the names of its contributors   #
# may be used to endorse or promote products derived from this software without   #
# specific prior written permission.                                              #
#                                                                                 #
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND #
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED   #
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE          #
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE    #
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL      #
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR      #
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER      #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,   #
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   #
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.            #
#                                                                                 #
###################################################################################


"""Global Configuration state and management"""


import getpass
import logging
import os

from contextlib import contextmanager

logger = logging.getLogger(__name__)


DEFAULTS = {
    "user": getpass.getuser(),
    "prefix": os.path.realpath(os.path.join(os.curdir, "prefix")),
    "cache": "cache",
}
"""Default values of the configuration"""


DOC = {
    "user": "User name for operations that create, delete or edit objects",
    "prefix": "Directory containing BEAT objects",
    "cache": "Directory to use for data caching (relative to prefix)",
}
"""Documentation for configuration parameters"""

_global_config = DEFAULTS.copy()


def get_config():
    """Retrieve current values for configuration set by :func:`set_config`

    Returns
    -------
    config : dict
        Keys are parameter names that can be passed to :func:`set_config`.

    See Also
    --------
    config_context: Context manager for global configuration
    set_config: Set global configuration
    """
    return _global_config.copy()


def set_config(**kwargs):
    """Set global configuration

    Parameters
    ----------
    user : str, optional
        The username used when creating objects
    prefix : str, optional
        The path to the current prefix
    cache : str, optional
        The path to the current cache

    See Also
    --------
    config_context: Context manager for global configuration
    get_config: Retrieve current values of the global configuration
    """
    supported_keys = set(DEFAULTS.keys())
    set_keys = set(kwargs.keys())
    if not set_keys.issubset(supported_keys):
        raise ValueError(
            f"Only {supported_keys} are valid configurations. "
            f"Got these extra values: {set_keys - supported_keys}"
        )
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    _global_config.update(kwargs)
    # if a new prefix path is set, clear the prefix
    if "prefix" in kwargs:
        Prefix().clear()


@contextmanager
def config_context(**new_config):
    """Context manager for global configuration

    Parameters
    ----------
    user : str, optional
        The username used when creating objects
    prefix : str, optional
        The path to the current prefix
    cache : str, optional
        The path to the current cache

    Notes
    -----
    All settings, not just those presently modified, will be returned to
    their previous values when the context manager is exited. This is not
    thread-safe.

    See Also
    --------
    set_config: Set global configuration
    get_config: Retrieve current values of the global configuration
    """
    old_config = get_config().copy()
    # also backup prefix
    prefix = Prefix()
    old_prefix = prefix.copy()
    set_config(**new_config)

    try:
        yield
    finally:
        set_config(**old_config)
        prefix.clear()
        prefix.update(old_prefix)


# ----------------------------------------------------------


class Singleton(type):
    """A Singleton metaclass
    From: https://stackoverflow.com/a/6798042/1286165
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class Prefix(dict, metaclass=Singleton):
    pass


class PrefixMeta(type):
    # cache instances when __init__ is called
    def __call__(cls, *args, **kwargs):
        instance = super().__call__(*args, **kwargs)

        folder = f"{cls.asset_folder}/{instance.name}"
        prefix = Prefix()
        prefix[folder] = instance
        print(f"caching {cls.__name__}/{instance.name} in __init__ calls")

        return instance

    def __init__(cls, name, bases, clsdict):
        # cache instances when new is called
        if "new" in clsdict:

            def caching_new(*args, **kwargs):
                instance = clsdict["new"].__func__(cls, *args, **kwargs)
                folder = f"{cls.asset_folder}/{instance.name}"
                prefix = Prefix()
                prefix[folder] = instance
                print(f"caching {cls.__name__}/{instance.name} in new calls")
                return instance

            setattr(cls, "new", caching_new)

        # change the key of instance when its name is changed
        if "name" in clsdict:
            name_property = clsdict["name"]
            actual_fset = name_property.fset

            def updating_fset(self, value):
                print(f"updating name of {cls.__name__}/{self.name} in cached prefix")
                if self.name in cls:
                    del cls[self.name]
                actual_fset(self, value)
                cls[self.name] = self

            name_property = property(
                name_property.fget,
                updating_fset,
                name_property.fdel,
                name_property.__doc__,
            )
            setattr(cls, "name", name_property)

    def __contains__(cls, key):
        return f"{cls.asset_folder}/{key}" in Prefix()

    def __getitem__(cls, key):
        folder = f"{cls.asset_folder}/{key}"
        prefix = Prefix()

        # if cached, return it
        if folder in prefix:
            return prefix[folder]

        # otherwise, load from prefix
        asset = super().__call__(key)
        return asset

    def __setitem__(cls, key, value):
        folder = f"{cls.asset_folder}/{key}"
        prefix = Prefix()
        prefix[folder] = value

    def __delitem__(cls, key):
        folder = f"{cls.asset_folder}/{key}"
        prefix = Prefix()
        del prefix[folder]
