import io
import unittest

import numpy as np

from ..dataformat import DataFormat

coordinates = DataFormat.new(
    definition={"x": int, "y": int, "name": str},
    name="coordinates",
    description="coordinates in an image",
)

sizes = DataFormat.new(
    definition={"width": int, "height": int, "array": [0, float]}, name="sizes"
)
rectangles = DataFormat.new(
    definition={"coords": coordinates, "size": sizes}, name="rectangles"
)


class CoordinatesToSizes:
    def process(self, inputs, data_loaders, outputs):
        # converts coordinates to sizes
        coord = inputs["coordinates"]
        new_size = sizes.type().from_dict(
            dict(width=coord.x, height=coord.y, array=[coord.x, coord.y])
        )
        outputs["sizes"].write(new_size)
        return True


class DataFormatTest(unittest.TestCase):
    def test_dataformat_creation(self):
        def assert_hash(h, oracle):
            self.assertEqual(h, oracle)

        # test hash functionality
        assert_hash(
            coordinates.hash(),
            "196214412087efe517820a27b6442cfb9fc2a843e30bc33c38bc74db48ee2bf1",
        )

        def assert_data_roundtrip(data, data_format, oracle=None):
            if oracle is None:
                oracle = data
            beat_type = data_format.type
            beat_data = beat_type().from_dict(data)
            obtained_data = beat_data.as_dict()
            self.assertEqual(obtained_data, oracle)
            # test validate
            data_format.validate(data)
            # validate from string
            data_format.validate(beat_data.pack())
            # validate from file
            with io.BytesIO() as fd:
                beat_data.pack_into(fd)
                fd.seek(0)
                data_format.validate(fd)

        # test data creation
        data = dict(x=1, y=2, name="test")
        assert_data_roundtrip(data, coordinates)

        # test safe data conversion
        data = dict(x=np.int32(1), y=np.int32(2), name="test")
        assert_data_roundtrip(data, coordinates)

        # test unsafe data conversion
        with self.assertRaises(TypeError):
            coordinates.type().from_dict(dict(x=1.0, y=2.0, name="test"))

        # test extra data attributes
        with self.assertRaises(AttributeError):
            coordinates.type().from_dict(dict(x=1, y=2, name="test", extra="five"))

        # test missing data attributes
        with self.assertRaises(AttributeError):
            coordinates.type().from_dict(dict(x=1))

        # test hierarchy
        data = dict(
            coords=dict(x=1, y=2, name="test"), size=dict(width=3, height=4, array=[1])
        )
        assert_data_roundtrip(data, rectangles)

        # test when data is already in BEAT format
        data_ = dict(
            coords=coordinates.type().from_dict(dict(x=1, y=2, name="test")),
            size=sizes.type().from_dict(
                dict(width=3, height=4, array=np.asarray([1], dtype="int32"))
            ),
        )
        assert_data_roundtrip(data_, rectangles, oracle=data)

        # test hash functionality
        assert_hash(
            sizes.hash(),
            "afcc84a5ea2a6dcdcd531352691a6358d1bcbd6c28e4d41ce15328dd7e5f314e",
        )

        assert_hash(
            rectangles.hash(),
            "a09ebe6da8e4b812c7d1e0ba0a841e35c988280687960ca5a622ee142e690316",
        )

        # name cannot be invalid
        with self.assertRaises(ValueError):
            DataFormat.new({"value": int}, name=None)

        with self.assertRaises(ValueError):
            DataFormat.new({"value": int}, name="")

        # test name change
        df = DataFormat.new({"value": int}, name="int")
        old_name = df.name
        df.name = "int2"
        new_name = df.name
        self.assertTrue(old_name not in DataFormat)
        self.assertTrue(new_name in DataFormat, new_name)

        # test creation with init
        df = DataFormat({"value": int})
        self.assertTrue(len(df.errors) == 0, df.errors)


# class AlgorithmTest(unittest.TestCase):
#     """docstring for AlgorithmTest"""

#     def test_algorithm_creation(self):
#         alg = CoordinatesToSizes()
#         algorithm = Algorithm_(
#             alg,
#             {"main": {"inputs": ["coordinates"], "outputs": ["sizes"]}},
#             type="sequential",
#         )
